# calc.rb - Ruby expression parsing classes
#
# Sterling W. (Chip) Camden             January, 2006
#

# A unique class for exceptions so we can rescue them separately if we want to

class ExpressionException < RuntimeError
end

# Any token we might encounter in the expression

class ParseToken
public
  # values for parseAction (what to do when we encounter this token):
  PARSE_VALUE = 0       # A literal or an identifier (stack it)
  PARSE_OP = 1          # An operator
  PARSE_PUSH = 2        # Open parenthesis, for example
  PARSE_POP = 3         # Close parenthesis, for example
  attr_reader :symbol, :parseAction
private
  def initialize(symbol, parseAction)
    @symbol = symbol             # symbol used for parsing
    @parseAction = parseAction   # action when parsing
  end
end

# An operator (or function identifier)

class Operator < ParseToken
  attr_reader :proc, :minarg, :maxarg, :priorityProc
  attr_writer :priorityProc
private
  def initialize(symbol, minarg, maxarg, priority, &proc)
    super(symbol, ParseToken::PARSE_OP)
    @minarg = minarg            # minimum number of operands
    @maxarg = maxarg            # maximum number of operands
    @priority = priority        # operator precedence (0 = low .. 5 = high)
    @proc = proc                # code block for performing the operation
    @priorityProc = nil         # procedure for determining priority
                                # (we need this for operators that determine
                                #  priority based on context, like + and -
  end
public
  # Get priority of the operator
  def priority(operandPreceding = nil)
    # only determine priority once, possibly based on context
    if (!(@priority) && @priorityProc)         # do we have a proc for processing priority?
      @priority = @priorityProc.call(operandPreceding)     # call it
    end
    @priority                   # return it
  end
end

# The next few classes are really only useful for self-documenting code and ease of typing

# An operator that takes only one operand

class UnaryOperator < Operator
private
  def initialize(symbol, priority)
    super(symbol, 1, 1, priority)
  end
end

# An operator that takes two operands

class BinaryOperator < Operator
private
  def initialize(symbol, priority)
    super(symbol, 2, 2, priority)
  end
end

# An operator that can go either way...

class BinOrUnOperator < Operator
private
  def initialize(symbol, priority)
    super(symbol, 1, 2, priority)
  end
end

# A function has the lowest priority, because usually the arguments are
# enclosed in a push/pop sequence (e.g., parentheses)

class Function < Operator
private
  def initialize(name, minarg, maxarg)
    super(name, minarg, maxarg, 5)
  end
end

# OK, back to something we really need.
# Here's our representative of an open parenthesis (or equivalent)

class Pusher < ParseToken
  attr_reader :parent, :implied  # parent operand array
  attr_writer :parent, :implied  # implied means we created it via operator precedence
private
  def initialize(symbol)
    super(symbol, ParseToken::PARSE_PUSH)
    @implied = false             # by default it is a literal token
  end
public
  def value                      # Tried to evaluate one of these
    raise(ExpressionException, "Unrecognized expression")
  end
end

# The complementary close parenthesis (or equivalent)

class Popper < ParseToken
private
  def initialize(symbol)
    super(symbol, ParseToken::PARSE_POP)
  end
end

# A token that presumably references a variable of some sort
# This class demonstrates calling a code block from a saved reference to it.

class Identifier < ParseToken
private
  def initialize(name, proc)
    super(name, ParseToken::PARSE_VALUE)        # Note we use the name as the symbol
    @proc = proc                                # proc is a code block for evaluating the symbol
  end
public
  def value=(newvalue)                          # Assign to value
    @proc.call(self.symbol, newvalue)           # We pass newvalue to denote assignment
  end
  def value                                     # Get value
    @proc.call(self.symbol)                     # No newvalue passed here
  end
end

# A literal value

class Literal < ParseToken
  attr_reader :value                            # Just like an Identifier, we can get the
                                                # value using the 'value' property, even 
                                                # though the implementation has nothing in common.
                                                # The nice thing about Ruby is that these
                                                # two classes do not need to be derived
                                                # from an abstract parent class (less typing,
                                                # in every sense of the word).
private
  def initialize(value)
    super(value, ParseToken::PARSE_VALUE)       # Note we use the value itself as the symbol
    @value = value
  end
public
  # We could have left out the following method, but our version gives a better error
  # messsage than the default "method missing value=".
  def value=(newvalue)                          # Assign a value
    raise(ExpressionException, "Writing into a literal")        # Not!
  end
end

# A special Hash for the list of supported operators
# The user can append new operators, without worrying about how we need to handle them.

class OperatorList < Hash
  def <<(oper)                                  # No need for the user to know about the keys
    self[oper.symbol] = oper                    # We use the operator's symbol for the key, BTW
  end
end

# An expression processor.  This class can be derived to process many types of
# expressions.  The modules defined below can be included in a derived class to 
# support popular sets of operations, or you can define your own.
# The code block passed to the new method for this class is used for evaluating
# identifiers in the expression.  This allows the client code to supply/store
# variables in any way it wishes.

class ExpressionProcessor
  attr_reader :operators, :last_result
private
  def initialize(&proc)
    @proc = proc                   # code block for accessing values for identifiers
    @operators = OperatorList.new  # hash for operators
    @curr = nil                    # current operand stack
    @currop = nil                  # current operator
    @opstack = nil                 # stack of operators
    @last_result = nil             # result of last expression

    # Here's a nifty little trick.  We'll inspect our methods and call any method
    # whose name starts with "AddOperator_".  That allows a derived class to include
    # any of the operator modules defined below, and automatically get support
    # for the operators included in each module, without having to invoke the
    # methods explicitly.

    methods.each do |name|
        self.send(name) if (name =~ /^AddOperators_/)
      end
  end

  # Resolve the expression composed of the current operator (@currop) and the
  # current operand stack (@curr)

  def resolve
    if (@currop) then              # Do we have an operator?
      args = @curr[1..-1]          # omit the "pusher" (curr[0] is always a Pusher)
      
      # check argument count requirements

      if (@currop.minarg && (args.length < @currop.minarg))
        raise(ExpressionException, "Not enough arguments for operation '#{@currop.symbol}'\nMinimum: #{@currop.minarg}, encountered #{args.length}")
      end
      if (@currop.maxarg && (args.length > @currop.maxarg))
        raise(ExpressionException, "Too many arguments for operation '#{@currop.symbol}'\nMaximum: #{@currop.maxarg}, encountered #{args.length}")
      end

      # Call the operator's code block and store the result as element 1, trimming the other operands

      @curr[1..-1] = Literal.new(@currop.proc.call(*args))
      @currop = nil             # No more current operator
    end
    if (@curr[0].implied)       # pop off any implied Pusher
      val = @curr[1]            # grab the value
      @curr = @curr[0].parent   # pop back up to the parent operand stack
      @curr[-1] = val           # store the value over the reference to the stack we just popped
      @currop = @opstack.pop    # pop off the operator from the previous level
    end
    return @curr[-1]            # In either case, return the last value on the current operand stack
  end
    
  end

public
  # Evaluate an expression
  def eval(input=$_)            # default to current string if not passed
    stack = [Pusher.new(nil)]   # operand array (a stack of operands and/or operand stacks)
    @opstack = []               # operator stack
    @curr = stack               # current operand stack
    @currop = nil               # operator at current stack level

    # Tokenize as numeric or identifier or non-space/comma (presumably an operator or parentheses)
    input.scan(/[0-9.]+|\w+|[^\s\t\r\n\f,]/) do |token|
        t = nil
        if (token =~ /^[0-9]+$/) then          # integer literal
          t = Literal.new(token.to_i)
        elsif (token =~ /^[0-9.]+$/) then      # floating point literal
          t = Literal.new(token.to_f)
        elsif (t = @operators[token]) then     # a defined operator
        elsif (token =~ /^\w+$/)               # an identifier
          t = Identifier.new(token, @proc)
        else
          raise(ExpressionException, "Unrecognized character in expression: #{token}")
        end

        if (t)                                 # Did we create something?
          case t.parseAction                   # What are we to do with it?
            when ParseToken::PARSE_VALUE       # Literals and identifiers
              @curr.push(t)                    #  get pushed as operands

            when ParseToken::PARSE_OP          # Operators and functions
              t = t.clone                      # Must copy to establish priority independently

              # Get the priority (note our only context-based syntax element, a leading operand)
              prior = t.priority(opPreceding = @currop ? (@curr.length > 2) : (@curr.length > 1))

              # If we already have an operator with at least equal priority, resolve it
              while (@currop && (prior <= @currop.priority))
                resolve
              end
              if (@currop) then             # We found an operator of lower priority
                pusher = Pusher.new(nil)    # Group our new operation (implied parenthesis)
                pusher.implied = true
                pusher.parent = @curr       # back-link the stack
                if (opPreceding) then       # operand preceding?
                  new = [pusher, @curr[-1]] # move last operand to new operation
                  @curr[-1] = new           # link new stack to end of old one
                else
                  new = [pusher]
                  @curr.push(new)           # just push and go
                end
                @curr = new                 # save reference to new operand stack
                @opstack.push(@currop)      # push the previous operator
              end
              @currop = t                   # now we have a new operator

              # Resolve the operation if we already reached the maximum argument count
              resolve if (@currop.maxarg && (@curr.length > @currop.maxarg))

            when ParseToken::PARSE_PUSH     # Grouping (parenthesis equivalent)
              t = t.clone                   # Clone so we have a unique parent property
              t.parent = @curr              # Back-link to previous operand stack
              @curr.push(new = [t])         # Forward-link to new operand stack
              @curr = new                   # Save reference to new stack
              @opstack.push(@currop)        # Push the current operator, if any
              @currop = nil                 # No current operator yet within this grouping

            when ParseToken::PARSE_POP      # End grouping (close parenthesis)
              if (@currop) then             # Do we have an operator?
                val = resolve               # Resolve the contained operation

                # If we're already unwound, this is an extra one
                raise(ExpressionException, "Unmatched #{t.symbol}") if (@curr == stack)
                @curr = @curr[0].parent     # Pop back to parent stack
                @curr[-1] = val             # Store the value over the forward reference (losing the reference)
              else                          # No current operator, just move the operands back
                val = @curr[1..-1]          # Gather the operands
                @curr = @curr[0].parent     # Pop back to parent stack
                @curr[-1,val.length] = val  # Extend the operand stack with the new operands
              end
              @currop = @opstack.pop        # Pop back to previous operator
          end                               # case t.parseAction
        end                                 # if (t)
      end                                   # input.scan

    # Pop off any implied pushes, resolving the values
    resolve while (@curr[0].implied)

    # Make sure we unwound completely
    raise(ExpressionException, "Unmatched #{@curr[0].symbol}") if (@curr != stack)

    # Resolve any final operator, and save the value
    @last_result = resolve.value

    # Make sure we ended up with just one value
    raise(ExpressionException, "Missing operator") if (@curr.length > 2)

    # Return that value
    @last_result
end

# Mix-ins for the usual operations.  See ExpressionProcessor#initialize for how
# these "AddOperator_" methods get invoked.

# First, parenthetical grouping

module ParentheticalGrouping
  def AddOperators_ParentheticalGrouping

    # Define "(" and ")" as our grouping operators

    @operators << Pusher.new("(")
    @operators << Popper.new(")")
  end
end

# Arithmetic operators + - * / % ^

module ArithmeticOperators
  def AddOperators_Arithmetic

    # Plus and minus are binary operators if they are preceeded by an operand.
    # Otherwise they are unary, and have higher priority.

    @operators << (plus = (BinOrUnOperator.new("+", nil) { |x, y| y ? x.value + y.value : x.value }))
    @operators << (minus = (BinOrUnOperator.new("-", nil) { |x, y| y ? x.value - y.value : -x.value }))

    # If a + or - has an operand preceding, the priority is lower than * and /,
    # but higher than =.  If no operand precedes, it is higher than * or /.
    # So, we create a code block to make this determination.

    pmPriority = Proc.new { |operandPreceding| operandPreceding ? 2 : 4 }

    # and assign it as the priorityProc for the + and - operators

    plus.priorityProc = pmPriority
    minus.priorityProc = pmPriority

    # Now we add some easy binary operators...

    @operators << (BinaryOperator.new("*", 3) { |x, y| x.value * y.value })
    @operators << (BinaryOperator.new("/", 3) { |x, y| x.value / y.value })
    @operators << (BinaryOperator.new("%", 3) { |x, y| x.value % y.value })
    @operators << (BinaryOperator.new("^", 3) { |x, y| x.value ** y.value })
  end
end

# = for assignment

module AssignmentOperator
  def AddOperators_Assignment
    @operators << (BinaryOperator.new("=", 1) { |x, y| x.value = y.value })
  end
end

# $ for last result

module LastResultOperator
  def AddOperators_LastResult
    @operators << Function.new("$", 0, 0) { self.last_result }
  end
end

# Boolean and or eq ne gt lt ge le not true false

module BooleanOperators
  def AddOperators_Boolean

    # Functions for true and false

    @operators << Function.new("true", 0, 0) { true }
    @operators << Function.new("false", 0, 0) { false }

    # Equality and its clan (note we cannot use '==' or other two-character
    # non-word operators, because of the way we parse the string.  Non-word
    # characters come in one at a time.

    @operators << BinaryOperator.new("eq", 3) { |x, y| x.value == y.value }
    @operators << BinaryOperator.new("ne", 3) { |x, y| x.value != y.value }
    @operators << BinaryOperator.new("gt", 3) { |x, y| x.value > y.value }
    @operators << BinaryOperator.new("lt", 3) { |x, y| x.value < y.value }
    @operators << BinaryOperator.new("ge", 3) { |x, y| x.value >= y.value }
    @operators << BinaryOperator.new("le", 3) { |x, y| x.value <= y.value }

    # Boolean and/or

    @operators << BinaryOperator.new("and", 2) { |x, y| x.value && y.value }
    @operators << BinaryOperator.new("or", 2) { |x, y| x.value || y.value }

    # Logical not

    @operators << UnaryOperator.new("not", 4) { |x| !x.value }

  end
end
