# Scotty TNG - ping/snmp based realtime network monitoring
#
# $Id: ups.rb 115 2008-10-02 13:03:24Z liske $
#
# Authors:
#   Thomas Liske <liske@ibh.de>
#   Andre Ellguth <ellguth@ibh.de>
#
# Copyright Holder:
#   2008 (C) IBH IT-Service GmbH [http://www.ibh.de/]
#
# License:
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this package; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA
#

module Sensors

  # This class implements a SNMP based UPS +Sensor+.
  class UPS < Sensor
    @@sensors = Array.new
    
    attr_reader :upsBatteryStatus_OID, :upsSecondsOnBattery_OID,
      :upsEstimatedMinutesRemaining_OID, :upsAlarmsPresent_OID,
      :upsOutputSource_OID, :mon_OID,
      :rlname, :model, :lines
    
    attr_accessor :mon_batt_stat, :mon_batt_onsec, :mon_batt_left,
      :mon_alarms, :mon_osrc, :mon_oload, :mon_reachable

    def initialize(node, xmlsensor)
      @@sensors.push(self)
      
      @rlname = "ups"
      @recs = Array.new
      @stamps = Array.new

      @mon_batt_stat = 0
      @mon_batt_onsec = 0
      @mon_batt_left = 0
      @mon_alarms = 0
      @mon_osrc = 0
      @mon_oload = 0
      @mon_reachable = 0

      @upsIdentModel_OID = "1.3.6.1.2.1.33.1.1.2.0"
      @sysLocation_OID = "1.3.6.1.2.1.1.6.0"
#      @upsIdentName_OID = "1.3.6.1.2.1.33.1.1.5.0"
#      @upsIdentAttachedDevices_OID = "1.3.6.1.2.1.33.1.1.6.0"
      @upsBatteryStatus_OID = "1.3.6.1.2.1.33.1.2.1.0"
      @upsSecondsOnBattery_OID = "1.3.6.1.2.1.33.1.2.2.0"
      @upsOutputSource_OID = "1.3.6.1.2.1.33.1.4.1.0"
      @upsOutputNumLines_OID = "1.3.6.1.2.1.33.1.4.3.0"
      @upsOutputPercentLoad_OID = "1.3.6.1.2.1.33.1.4.4.1.5.#"
      @upsEstimatedMinutesRemaining_OID = "1.3.6.1.2.1.33.1.2.3.0"
      @upsAlarmsPresent_OID = "1.3.6.1.2.1.33.1.6.1.0"

      @mon_OID = [@upsBatteryStatus_OID, @upsSecondsOnBattery_OID, @upsEstimatedMinutesRemaining_OID,
		  @upsAlarmsPresent_OID, @upsOutputSource_OID]

      super(node, xmlsensor, @rlname)

      begin
	response = node.snmp.get([@upsIdentModel_OID, @sysLocation_OID, @upsOutputNumLines_OID])
      rescue
	puts "\nERROR: Could not query SNMP: #{$!}"
	exit 1
      end

      model = nil
      location = nil
      @lines = nil

      response.each_varbind do |vb|
	begin
	    case vb.name.to_s
	    when @upsIdentModel_OID
		model = vb.value.to_s.strip
	    when @sysLocation_OID
		location = vb.value.to_s.strip
	    when @upsOutputNumLines_OID
		@lines = vb.value.to_i
	    end
	rescue
	    puts "Error retrieving #{vb.name.to_s}: #{$!}"
	end
      end

      max = [0, 0, 0, 0, 0]
      format = ["ups_bat", "t", "t", "c", "ups_src"]
      type = ["v", "v", "v", "v", "v"]
      color = ["Purple", "Blue", "Green", "Red", "Orange"]
      legend = ["Battery status", "On battery", "Battery left", "Alarms present", "Output source"]
      slegend = [nil, nil, "l", nil, nil]
      1.upto(@lines) {|l|
        max.push(100)
        format.push("p")
        type.push("v")
        color.push("Black")
        legend.push("Output ##{l} load")
        slegend.push(l.to_s)

        @mon_OID.push(@upsOutputPercentLoad_OID.sub(/#/, l.to_s))
      }

      RingList.new(node, self, @rlname , {
		     :lines => max.size,
		     :title => model,
		     :descr => location,
		     :class => @class,
		     :max => max,
		     :format => format,
		     :type => type,
		     :color => color,
		     :legend => legend,
		     :slegend => slegend,
		   })

      self.add_expr("status") { @mon_batt_stat }
      self.add_expr("onbat") { @mon_batt_onsec }
      self.add_expr("tleft") { @mon_batt_left }
      self.add_expr("alarms") { @mon_alarms }
      self.add_expr("osrc") { @mon_osrc }
      self.add_expr("oload") { @mon_oload }
      self.add_expr("reachable") { @mon_reachable }
    end

    def self.monitor()
      while true do
	@@sensors.each{ |s|
	    stat = 0
	    onsec = 0
	    left = 0
	    alarms = 0
	    osrc = 1
	    oload = Array.new
	    reachable = 0

	    response = s.node.snmp.get(s.mon_OID)
	    response.each_varbind do |vb|
	      begin
	      vname = vb.name.to_s
	        case vb.name.to_s
	        when s.upsBatteryStatus_OID
		  stat = vb.value.to_i
	        when s.upsSecondsOnBattery_OID
		  onsec = vb.value.to_i * 1000
	        when s.upsEstimatedMinutesRemaining_OID
		  left = vb.value.to_i * 60000
	        when s.upsAlarmsPresent_OID
		  alarms = vb.value.to_i
	        when s.upsOutputSource_OID
		  osrc = vb.value.to_i
		else
		  oload.push(vb.value.to_i)
	        end

	        reachable = 1
	      rescue
		print "#{s.node.name}/#{s.rlname}/#{vb.name}: "
		p $!
	      end
	    end

	    s.mon_batt_stat = stat
	    s.mon_batt_onsec = onsec
	    s.mon_batt_left = left
	    s.mon_alarms = alarms
	    s.mon_osrc = osrc
	    s.mon_oload = oload.sort.pop
	    s.mon_reachable = reachable

	    s.check_mon

	    s.pushProbe([
						   s.mon_batt_stat,
						   s.mon_batt_onsec,
						   s.mon_batt_left,
						   s.mon_alarms,
						   s.mon_osrc
						 ].concat(oload))
	}
	
	sleep 5
      end
    end
  end
end
